/*
 *
 *	Utility to check disk quotas
 *
 *	Some parts of this utility are copied from old quotacheck by
 *	Marco van Wieringen <mvw@planets.elm.net> and Edvard Tuinder <ed@elm.ent>
 * 
 */

#include <dirent.h>
#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include <getopt.h>
#include <limits.h>
#include <unistd.h>
#include <stdlib.h>
#include <errno.h>

#include <asm/types.h>
#include <asm/byteorder.h>

#include <linux/quota.h>

#include <sys/stat.h>
#include <sys/types.h>
#include <sys/param.h>
#include <sys/file.h>
#include <sys/statfs.h>
#include <sys/mount.h>

#if defined(EXT2_DIRECT)
#include <linux/ext2_fs.h>
#include <ext2fs/ext2fs.h>
#endif

#include "common.h"
#include "quota_io.h"
#include "quotasys.h"

#define NODQUOT (struct check_dquot *)NULL

#define LINKSHASHSIZE 16384	/* Size of hashtable for hardlinked inodes */
#define DQUOTHASHSIZE 32768	/* Size of hashtable for dquots from file */

struct check_dquot {
	qid_t dq_id;  /* id this applies to (uid, gid) */
	struct mem_dqblk dq_dqb; /* diskquota for id */
	struct check_dquot *next;  /* pointer to next id */
};

struct dlinks {
	ino_t i_num;
	struct dlinks *next;
};

struct dirs {
	char *dir_name;
	struct dirs *next;
};

#define BITS_SIZE 4 /* sizeof(bits) == 5 */

#define FL_FORCE 1		/* Force check even if quota enabled */
#define FL_VERBOSE 2		/* Have verbose input */
#define FL_DEBUG 4		/* Have very verbose input */
#define FL_INTERACTIVE 8	/* Ask questions when needed */
#define FL_GUESSDQ 16		/* When more structures for same user found, use the first */
#define FL_NEWFILE 32		/* Don't try to read old file. Just create new one. */
#define FL_FORCEREMOUNT 64	/* Force check even when remounting RO fails */
#define FL_NOREMOUNT 128	/* Don't try to remount filesystem RO */

#define SET_BLK(blk) (blkbmp[(blk) >> 3] |= 1 << ((blk) & 7))
#define GET_BLK(blk) (blkbmp[(blk) >> 3] & (1 << ((blk) & 7)))

const char bits[] = "|/-\\";
const int magics[MAXQUOTAS] = INITQMAGICS;	/* Magics we should look for */
const int known_versions[MAXQUOTAS] = INITKNOWNVERSIONS;	/* Versions we accept */
dev_t cur_dev;			/* Device we are working on */
int files_done, dirs_done;	
int flags, type, check_type;	/* Options from command line; Types we want to check; Type we are checking */
char *mntpoint;			/* Mountpoint to check */
struct mem_dqinfo old_info[MAXQUOTAS];	/* Loaded infos */
#ifdef DEBUG_MALLOC
static size_t malloc_mem = 0;
static size_t free_mem = 0;
#endif
char *blkbmp;

struct check_dquot *dquot_hash[MAXQUOTAS][DQUOTHASHSIZE];
struct dlinks *links_hash[MAXQUOTAS][DQUOTHASHSIZE];

/*
 * Ok check each memory allocation.
 */
void *xmalloc(size_t size)
{
	void *ptr;

#ifdef DEBUG_MALLOC
	malloc_mem += size;
#endif
	ptr = malloc(size);
	if (!ptr)
		die(3, "Not enough memory.\n");
	memset(ptr, 0, size);
	return(ptr);
}

void debug(int df, char *fmtstr, ...)
{
	va_list args;

	if (!(flags & df))
		return;
	va_start(args, fmtstr);
	vfprintf(stderr, fmtstr, args);
	va_end(args);	  
}

/* Compute hashvalue for given inode number */
inline uint hash_ino(uint i_num)
{
	return ((i_num ^ (i_num << 16)) * 997) & (LINKSHASHSIZE - 1);
}

/*
 * Store a hardlinked inode as we don't want to count it more then once.
 */
int store_dlinks(int type, ino_t i_num)
{
	struct dlinks *lptr;
	int kt = user2kernquota(type);
	uint hash = hash_ino(i_num);

	debug(FL_DEBUG, "Adding hardlink for ino %d\n", i_num);

	for (lptr = links_hash[kt][hash]; lptr; lptr = lptr->next)
		if (lptr->i_num == i_num)
			return 1;

	lptr = (struct dlinks *)xmalloc(sizeof(struct dlinks));
	lptr->i_num = i_num;
	lptr->next = links_hash[kt][hash];
	links_hash[kt][hash] = lptr;
	return 0;
}

/* Hash given id */
inline uint hash_dquot(uint id)
{
	return ((id ^ (id << 16)) * 997) & (DQUOTHASHSIZE - 1);
}

/*
 * Do a lookup of a type of quota for a specific id. Use short cut with
 * most recently used dquot struct pointer.
 */
struct check_dquot *lookup_dquot(qid_t id, int type)
{
	struct check_dquot *lptr;
	int kt = user2kernquota(type);
	uint hash = hash_dquot(id);

	for (lptr = dquot_hash[kt][hash]; lptr != NODQUOT; lptr = lptr->next)
		if (lptr->dq_id == id)
			return lptr;
	return NODQUOT;
}

/*
 * Add a new dquot for a new id to the list.
 */
inline struct check_dquot *add_dquot(qid_t id, int type)
{
	struct check_dquot *lptr;
	int kt = user2kernquota(type);
	uint hash = hash_dquot(id);

	debug(FL_DEBUG, "Adding dquot structure type %s for %d\n", type2name(type), (int)id);

	lptr = (struct check_dquot *)xmalloc(sizeof(struct check_dquot));

	lptr->dq_id = id;
	lptr->next = dquot_hash[kt][hash];
	dquot_hash[kt][hash] = lptr;
	lptr->dq_dqb.dqb_btime = lptr->dq_dqb.dqb_itime = (time_t) 0;

	return lptr;
}

/*
 * Add a number of blocks and inodes to a quota.
 */
void add_to_quota(int type, ino_t i_num, uid_t i_uid, gid_t i_gid, umode_t i_mode,
						nlink_t i_nlink, uint i_blocks)
{
	qid_t wanted;
	struct check_dquot *lptr;

	if (type & QS_USRQUOTA)
		wanted = i_uid;
	else
		wanted = i_gid;

	if ((lptr = lookup_dquot(wanted, type)) == NODQUOT)
		lptr = add_dquot(wanted, type);

	if (i_nlink != 1)
		if (store_dlinks(type, i_num))	/* Did we already count this inode? */
			return;
	lptr->dq_dqb.dqb_curinodes++;
	if (i_blocks)
		lptr->dq_dqb.dqb_curspace += ((__u64)i_blocks)*(QUOTABLOCK_SIZE>>1);
}

/*
 * Clean up all list from a previous run.
 */
void remove_list(void)
{
	int cnt;
	uint i;
	struct check_dquot *dquot, *dquot_free;
	struct dlinks *dlink, *dlink_free;

	for (cnt = 0; cnt < MAXQUOTAS; cnt++) {
		for (i = 0; i < DQUOTHASHSIZE; i++) {
			dquot = dquot_hash[cnt][i];
			while (dquot != NODQUOT) {
				dquot_free = dquot;
				dquot = dquot->next;
#ifdef DEBUG_MALLOC
		 free_mem += sizeof(struct dquot);
#endif
				free(dquot_free);
			}
			dquot_hash[cnt][i] = NODQUOT;
		}
		for (i = 0; i < LINKSHASHSIZE; i++) {
			dlink = links_hash[cnt][i];
			while (dlink) {
				dlink_free = dlink;
				dlink = dlink->next;
#ifdef DEBUG_MALLOC
		 free_mem += sizeof(struct dlinks);
#endif
				free(dlink_free);
			}
			links_hash[cnt][i] = NULL;
		}
	}
}

/*
 * Show a blitting cursor as means of visual progress indicator.
 */
static inline void blit(void)
{
	static short bitc = 0;

	putc(bits[bitc], stdout);
	putc('\b', stdout);
	fflush(stdout);
	bitc++;
	bitc %= BITS_SIZE;
}

void parse_options(int argcnt, char **argstr)
{
	int ret;
	char *slash = strrchr(argstr[0], '/');

	if (slash)
		slash++;
	else
		slash = argstr[0];

	while ((ret = getopt(argcnt, argstr, "VhcvugidnfFR")) != EOF) {
		switch (ret) {
			case 'g':
				type |= QS_GRPQUOTA;
				break;
			case 'u':
				type |= QS_USRQUOTA;
				break;
			case 'd':
				flags |= FL_DEBUG;
				setlinebuf(stderr);
				 break;
			case 'v':
				flags |= FL_VERBOSE;
				break;
			case 'f':
				flags |= FL_FORCE;
				break;
			case 'i':
				flags |= FL_INTERACTIVE;
				break;
			case 'n':
				flags |= FL_GUESSDQ;
				break;
			case 'c':
				flags |= FL_NEWFILE;
				break;
			case 'V':
				version();
				exit(0);
			case 'F':
				flags |= FL_FORCEREMOUNT;
				break;
			case 'R':
				flags |= FL_NOREMOUNT;
				break;
			default:
usage:
				printf("Utility for checking and repairing quota files.\n%s [-gucfinvdFR] filesystem\n", slash);
				printf("Bugs to %s\n", MY_EMAIL);
				exit(1);
		}
	}
	if (!(type & (QS_USRQUOTA | QS_GRPQUOTA)))
		type |= QS_USRQUOTA;
	if (argcnt == optind) {
		fputs("Bad number of arguments.\n", stderr);
		goto usage;
	}
	if (flags & FL_VERBOSE && flags & FL_DEBUG)
		flags &= ~FL_VERBOSE;
	mntpoint = argstr[optind];
}

#if defined(EXT2_DIRECT)
void ext2_direct_scan(char *device)
{
	ino_t i_num;
	ext2_filsys fs;
	errcode_t error;
	ext2_inode_scan scan;
	struct ext2_inode inode;
	int inode_buffer_blocks = 0;
	ext2fs_inode_bitmap inode_used_map;
	ext2fs_inode_bitmap inode_dir_map;

	if ((error = ext2fs_open(device, 0, 0, 0, unix_io_manager, &fs)))
		die(2, "quotacheck: error (%d) while opening %s\n", error, device);

	if ((error = ext2fs_allocate_inode_bitmap(fs, "in-use inode map", &inode_used_map)))
		die(2, "quotacheck: error (%d) while allocating inode file bitmap\n", error);

	if ((error = ext2fs_allocate_inode_bitmap(fs, "directory inode map", &inode_dir_map)))
		die(2, "quotacheck: error (%d) while allocating inode directory bitmap\n", error);

	if ((error = ext2fs_open_inode_scan(fs, inode_buffer_blocks, &scan)))
		die(2, "quotacheck: error (%d) while opening inode scan\n", error);

	if ((error = ext2fs_get_next_inode(scan, &i_num, &inode)))
		die(2, "quotacheck: error (%d) while starting inode scan\n", error);

	while (i_num) {
		if (inode.i_links_count) {
			debug(FL_DEBUG, "Found i_num %ld\n", i_num);
			if (flags & FL_VERBOSE)
				blit();
			if (check_type & QS_USRQUOTA)
				 add_to_quota(QS_USRQUOTA, i_num, inode.i_uid, inode.i_gid, inode.i_mode,
								 inode.i_links_count, inode.i_blocks);
			if (check_type & QS_GRPQUOTA)
				 add_to_quota(QS_GRPQUOTA, i_num, inode.i_uid, inode.i_gid, inode.i_mode,
								 inode.i_links_count, inode.i_blocks);
			if (S_ISDIR(inode.i_mode))
				 dirs_done++;
			else
				 files_done++;
		}

		if ((error = ext2fs_get_next_inode(scan, &i_num, &inode)))
			die(2, "Something weird happened while scanning. Error %d\n", error);
	}
}
#endif

/*
 * Scan a directory with the readdir systemcall. Stat the files and add the sizes
 * of the files to the appropriate quotas. When we find a dir we recursivly call
 * ourself to scan that dir.
 */
void scan_dir(char *pathname)
{
	struct dirs *dir_stack = {(struct dirs *)NULL};
	struct dirs *new_dir;
	struct dirent *de;
	struct stat st;
	DIR *dp;

	if ((dp = opendir(pathname)) == (DIR *)NULL)
		die(2, "\nCan open directory %s: %s\n", pathname, strerror(errno));

	chdir(pathname);
	while ((de = readdir(dp)) != (struct dirent *)NULL) {
		if (!strcmp(de->d_name, ".") || !strcmp(de->d_name, ".."))
			continue;
		if (flags & FL_VERBOSE)
			blit();

		if ((lstat(de->d_name, &st)) == -1)
	 		die(2, "lstat can't stat `%s/%s': %s\nGuess you'd better run fsck first !\nexiting...\n", pathname, de->d_name, strerror(errno));

		if (check_type & QS_USRQUOTA)
			add_to_quota(QS_USRQUOTA, st.st_ino, st.st_uid, st.st_gid, st.st_mode,
							 st.st_nlink, st.st_blocks);
		if (check_type & QS_GRPQUOTA)
			add_to_quota(QS_GRPQUOTA, st.st_ino, st.st_uid, st.st_gid, st.st_mode,
							 st.st_nlink, st.st_blocks);

		if (S_ISDIR(st.st_mode)) {
			if (st.st_dev != cur_dev)
				continue;
			/*
			 * Add this to the directory stack and check this later on.
			 */
	 		debug(FL_DEBUG, "pushd %s/%s\n", pathname, de->d_name);
			new_dir = xmalloc(sizeof(struct dirs));
			new_dir->dir_name = xmalloc(strlen(pathname) + strlen(de->d_name) + 2);
			sprintf(new_dir->dir_name, "%s/%s", pathname, de->d_name);
			new_dir->next = dir_stack;
			dir_stack = new_dir;
		} else {
			debug(FL_DEBUG, "\tAdding %s size %d ino %d links %d\n", de->d_name, st.st_size, st.st_ino, st.st_nlink);
			files_done++;
		}
	}
	closedir(dp);

	/*
	 * Traverse the directory stack, and check it.
	 */
	debug(FL_DEBUG, "Scanning stored directories from directory stack\n");
	while (dir_stack != (struct dirs *)NULL) {
		new_dir = dir_stack;
		dir_stack = dir_stack->next;
		debug(FL_DEBUG, "popd %s\nEntering directory %s\n", new_dir->dir_name, new_dir->dir_name);
		scan_dir(new_dir->dir_name);
		dirs_done++;
#ifdef DEBUG_MALLOC
		free_mem += sizeof(struct dirs) + strlen(new_dir->dir_name) + 1;
#endif
		free(new_dir->dir_name);
		free(new_dir);
	}
	debug(FL_DEBUG, "Leaving %s\n", pathname);
}

/* Ask user y/n question */
int ask_yn(char *q, int def)
{
	char a[10];	/* Users answer */

	printf("%s [%c]: ", q, def?'y':'n');
	fflush(stdout);
	while (1) {
		fgets(a, sizeof(a), stdin);
		if (*a == '\n')
			return def;
		if (!strcasecmp(a, "y\n"))
			return 1;
		if (!strcasecmp(a, "n\n"))
			return 0;
		printf("Illegal answer. Please answer y/n: ");
		fflush(stdout);
	}
}

int check_blkref(uint blk, uint blocks)
{
	if (blk >= blocks)
		return 0;
	if (blk && blk < DQTREEOFF)
		return 0;
	return 1;
}

/* Load and check basic info about quotas */
void check_info(char *filename, int fd, int type)
{
	struct disk_dqinfo dinfo;
	uint blocks, dflags, freeblk, freeent;
	off_t filesize;
	int kt = user2kernquota(type), err;

	lseek(fd, DQINFOOFF, SEEK_SET);
	err = read(fd, &dinfo, sizeof(struct disk_dqinfo));
	if (err < 0)
		die(2, "Can't read info from quota file %s: %s\n", filename, strerror(errno));
	if (err != sizeof(struct disk_dqinfo)) {
		fprintf(stderr, "WARNING: Quota file %s was probably truncated. Can't save quota settings...\n", filename);
		return;
	}

	blocks = __le32_to_cpu(dinfo.dqi_blocks);
	freeblk = __le32_to_cpu(dinfo.dqi_free_blk);
	freeent = __le32_to_cpu(dinfo.dqi_free_entry);
	dflags = __le32_to_cpu(dinfo.dqi_flags);
	filesize = lseek(fd, 0, SEEK_END);
	if (!check_blkref(freeblk, blocks) || dflags & ~DQF_MASK ||
	    !check_blkref(freeent, blocks) || (filesize + DQBLKSIZE - 1) >> DQBLKSIZE_BITS != blocks) {
		fprintf(stderr, "WARNING: Quota file info was corrupted.\n");
		debug(FL_DEBUG, "Size of file: %lu\nBlocks: %u Free block: %u Block with free entry: %u Flags: %x\n", (unsigned long)filesize, blocks, freeblk, freeent, dflags);
		old_info[kt].dqi_bgrace = MAX_DQ_TIME;
		old_info[kt].dqi_igrace = MAX_IQ_TIME;
		old_info[kt].dqi_blocks = (filesize + DQBLKSIZE - 1) >> DQBLKSIZE_BITS;
		old_info[kt].dqi_flags = 0;
		printf("Setting grace times and other flags to default values.\nAssuming number of blocks is %u.\n", old_info[kt].dqi_blocks);
	}
	else {
		old_info[kt].dqi_bgrace = __le32_to_cpu(dinfo.dqi_bgrace);
		old_info[kt].dqi_igrace = __le32_to_cpu(dinfo.dqi_igrace);
		old_info[kt].dqi_blocks = blocks;
		old_info[kt].dqi_flags = dflags;
	}
	old_info[kt].dqi_free_blk = old_info[kt].dqi_free_entry = 0;	/* This won't be needed */
}

/* Print error message */
void blk_corrupted(int *corrupted, uint *lblk, uint blk, char *fmtstr, ...)
{
	va_list args;

	if (!*corrupted) {
		if (!(flags & (FL_VERBOSE | FL_DEBUG)))
			fprintf(stderr, "Corrupted blocks: ");
	}
	if (flags & (FL_VERBOSE | FL_DEBUG)) {
	  va_start(args, fmtstr);
	  fprintf(stderr, "Block %u: ", blk);
	  vfprintf(stderr, fmtstr, args);
	  fputc('\n', stderr);
	  va_end(args);
	}
	else
	  if (*lblk != blk)
		 if (!*corrupted)
			 fprintf(stderr, "%u", blk);
		 else
			 fprintf(stderr, ", %u", blk);
	*corrupted = 1;
	*lblk = blk;
	fflush(stderr);
}

/* Put one entry info memory */
void buffer_entry(dqbuf_t buf, uint blk, int *corrupted, uint *lblk, int cnt, int type)
{
	struct mem_dqblk mdq, *fdq;
	qid_t id;
	struct check_dquot *cd;

	disk2memdqb(&mdq, ((struct disk_dqblk *)(((char *)buf) + sizeof(struct disk_dqdbheader))) + cnt);
	id = __le32_to_cpu(((struct disk_dqblk *)(((char *)buf) + sizeof(struct disk_dqdbheader)))[cnt].dqb_id);
	cd = lookup_dquot(id, type);
	if (cd != NODQUOT) {
		fdq = &cd->dq_dqb;
		if (mdq.dqb_bhardlimit != fdq->dqb_bhardlimit || mdq.dqb_bsoftlimit != fdq->dqb_bsoftlimit ||
		    mdq.dqb_ihardlimit != fdq->dqb_ihardlimit || mdq.dqb_isoftlimit != fdq->dqb_isoftlimit) {
			blk_corrupted(corrupted, lblk, blk, "Duplicated entries.");
			if (flags & FL_GUESSDQ) {
				if (!(flags & (FL_DEBUG | FL_VERBOSE)))
					fputc('\n', stderr);
				fprintf(stderr, "Found more structures for ID %u. Using values: BHARD: %u BSOFT: %u IHARD: %u ISOFT: %u\n", (uint)id,
				  fdq->dqb_bhardlimit, fdq->dqb_bsoftlimit, fdq->dqb_ihardlimit, fdq->dqb_isoftlimit);
				return;
			}
			else if (flags & FL_INTERACTIVE) {
				fprintf(stderr, "\nFound more structures for ID %u. Values: BHARD: %u/%u BSOFT: %u/%u IHARD: %u/%u ISOFT: %u/%u\n", (uint)id,
				  fdq->dqb_bhardlimit, mdq.dqb_bhardlimit, fdq->dqb_bsoftlimit, mdq.dqb_bsoftlimit,
				  fdq->dqb_ihardlimit, mdq.dqb_ihardlimit, fdq->dqb_isoftlimit, mdq.dqb_isoftlimit);
				if (ask_yn("Should I use new values", 0)) {
					fdq->dqb_bhardlimit = mdq.dqb_bhardlimit;
					fdq->dqb_bsoftlimit = mdq.dqb_bsoftlimit;
					fdq->dqb_ihardlimit = mdq.dqb_ihardlimit;
					fdq->dqb_isoftlimit = mdq.dqb_isoftlimit;
					fdq->dqb_btime = mdq.dqb_btime;
					fdq->dqb_itime = mdq.dqb_itime;
				}
			}
			else
				die(7, "ID %u has more structures. User intervention needed (use -i for interactive mode or -n for automatic answer).\n", (uint)id);
		}
		else if (mdq.dqb_itime != fdq->dqb_itime || mdq.dqb_btime != fdq->dqb_btime) {
			if (fdq->dqb_btime < mdq.dqb_btime)
				fdq->dqb_btime = mdq.dqb_btime;
			if (fdq->dqb_itime < mdq.dqb_itime)
				fdq->dqb_itime = mdq.dqb_itime;
		}
	}
	else {
		cd = add_dquot(id, type);
		fdq = &cd->dq_dqb;
		fdq->dqb_bhardlimit = mdq.dqb_bhardlimit;
		fdq->dqb_bsoftlimit = mdq.dqb_bsoftlimit;
		fdq->dqb_ihardlimit = mdq.dqb_ihardlimit;
		fdq->dqb_isoftlimit = mdq.dqb_isoftlimit;
		fdq->dqb_btime = mdq.dqb_btime;
		fdq->dqb_itime = mdq.dqb_itime;
	}
}

void check_read_blk(int fd, uint blk, dqbuf_t buf)
{
	size_t rd;

	lseek(fd, blk << DQBLKSIZE_BITS, SEEK_SET);
	rd = read(fd, buf, DQBLKSIZE);
	if (rd < 0)
		die(2, "Can't read block %u: %s\n", blk, strerror(errno));
	if (rd != DQBLKSIZE) {
		debug(FL_VERBOSE | FL_DEBUG, "Block %u is truncated.\n", blk);
		memset(buf + rd, 0, DQBLKSIZE-rd);
	}
}

int check_tree_ref(uint blk, uint ref, uint blocks, int check_use, uint *corrupted, uint *lblk)
{
	if (!check_blkref(ref, blocks))
		blk_corrupted(corrupted, lblk, blk, "Reference to illegal block %u", ref);
	if (!ref)
		return 1;
	if (!check_use || !GET_BLK(ref))
		return 1;
	blk_corrupted(corrupted, lblk, blk, "Block %u in tree referenced twice", ref);
	return 0;
}

/* Check block with structures */
void check_data_blk(int fd, uint blk, int type, uint blocks, uint *corrupted, uint *lblk)
{
	dqbuf_t buf = getdqbuf();
	struct disk_dqdbheader *head = (struct disk_dqdbheader *)buf;
	int i;
	struct disk_dqblk *dd = (struct disk_dqblk *)(head + 1);

	SET_BLK(blk);
	check_read_blk(fd, blk, buf);
	if (!check_blkref(__le32_to_cpu(head->dqdh_next_free), blocks))
		blk_corrupted(corrupted, lblk, blk, "Illegal free block reference to block %u", __le32_to_cpu(head->dqdh_next_free));
	if (__le16_to_cpu(head->dqdh_entries) > DQSTRINBLK)
		blk_corrupted(corrupted, lblk, blk, "Corrupted number of used entries (%u)", (uint)__le16_to_cpu(head->dqdh_entries));
	for (i = 0; i < DQSTRINBLK; i++)
		if (!empty_dquot(dd+i))
			buffer_entry(buf, blk, corrupted, lblk, i, type);
	freedqbuf(buf);
}

/* Check one tree block */
void check_tree_blk(int fd, uint blk, int depth, int type, uint blocks, uint *corrupted, uint *lblk)
{
	dqbuf_t buf = getdqbuf();
	__u32 *r = (__u32 *)buf;
	int i;

	SET_BLK(blk);
	check_read_blk(fd, blk, buf);
	for (i = 0; i < DQBLKSIZE >> 2; i++)
		if (depth < DQTREEDEPTH - 1) {
			if (check_tree_ref(blk, __le32_to_cpu(r[i]), blocks, 1, corrupted, lblk) && __le32_to_cpu(r[i]))	/* Isn't block OK? */
				check_tree_blk(fd, __le32_to_cpu(r[i]), depth+1, type, blocks, corrupted, lblk);
		}
		else if (check_tree_ref(blk, __le32_to_cpu(r[i]), blocks, 0, corrupted, lblk) && __le32_to_cpu(r[i]))
			check_data_blk(fd, __le32_to_cpu(r[i]), type, blocks, corrupted, lblk);
	freedqbuf(buf);
}

/* Load data from file to memory */
void buffer_file(int fd, int type)
{
	uint blocks = old_info[user2kernquota(type)].dqi_blocks, lastblk = 0;
	int corrupted = 0;

	blkbmp = xmalloc((blocks+7)>>3);
	memset(blkbmp, 0, (blocks+7)>>3);
	if (check_tree_ref(0, DQTREEOFF, blocks, 1, &corrupted, &lastblk))
		check_tree_blk(fd, DQTREEOFF, 0, type, blocks, &corrupted, &lastblk);
	else
		fprintf(stderr, "Can't gather quota data. Tree root node corrupted.\n");
	free(blkbmp);
	if (corrupted) {
		if (!(flags & (FL_VERBOSE | FL_DEBUG)))
			fputc('\n', stderr);
		fprintf(stderr, "WARNING: Some data might be changed due to corruption.\n");
	}
	else
		debug(FL_DEBUG | FL_VERBOSE, "Not found any corrupted blocks. Congratulations.\n");
}

/* Do checks and buffer quota file into memory */
void process_file(char *dev, char *mnt, int type)
{
	struct disk_dqheader head;
	char filename[MAXFILENAMELEN];
	int err, fd;

	debug(FL_DEBUG | FL_VERBOSE, "Going to check %s quota file of %s\n", type2name(type), mnt);

	if (!quotactl(QCMD(Q_GETINFO, user2kernquota(type)), dev, 0, (void *)old_info + user2kernquota(type))) { /* Is quota enabled? */
		if (!(flags & FL_FORCE))
			if (flags & FL_INTERACTIVE) {
				printf("Quota for %ss is enabled on mountpoint %s so quotacheck might damage the file.\n", type2name(type), mnt);
				if (!ask_yn("Should I continue", 0))
					die(6, "As you wish... Finishing.\n");
			}
			else
				die(6, "Quota for %ss is enabled on mountpoint %s so quotacheck might damage the file.\n\
Please turn quotas off or use -f to force checking.\n", type2name(type), mnt);
		/* At least sync quotas so damage will be smaller */
		if (quotactl(QCMD(Q_SYNC, user2kernquota(type)), dev, 0, NULL) < 0)
			die(4, "Error while syncing quotas: %s\n", strerror(errno));
	}
	else if (errno != ESRCH && errno != ENOSYS && errno != EINVAL)
		die(6, "Huh... Error while getting info about quotas: %s\nProbably you have the honour of finding a bug.\n", strerror(errno));

	get_qfname(mnt, type, filename);
	if (flags & FL_NEWFILE) {	/* Should we just create new file? */
		int kt = user2kernquota(type);

		old_info[kt].dqi_bgrace = MAX_DQ_TIME;
		old_info[kt].dqi_igrace = MAX_IQ_TIME;
		return;
	}
	if ((fd = open(filename, O_RDONLY)) < 0)
		die(1, "Can't open quota file %s: %s\n", filename, strerror(errno));
	err = read(fd, &head, sizeof(struct disk_dqheader));
	if (err < 0)
		die(1, "Can't read header from quota file %s: %s\n", filename, strerror(errno));
	if (err != sizeof(struct disk_dqheader)) {
		fprintf(stderr, "WARNING: Quota file %s was probably truncated. Can't save quota settings...\n", filename);
		return;
	}
	if (__le32_to_cpu(head.dqh_magic) != magics[user2kernquota(type)] || __le32_to_cpu(head.dqh_version) > known_versions[user2kernquota(type)])
		fprintf(stderr, "WARNING: Quota file %s has corrupted headers\n", filename);
	debug(FL_DEBUG, "Headers checked. Checking file info...\n");
	check_info(filename, fd, type);
	debug(FL_DEBUG, "File info done.\n");
	debug(FL_DEBUG | FL_VERBOSE, "File %s checked. Going to load data...\n", filename);
	/* Load data from file */
	buffer_file(fd, type);
	close(fd);
}

/* Backup old quotafile and rename new one to right name */
void rename_files(char *mnt, int type)
{
	char filename[MAXFILENAMELEN], newfilename[MAXFILENAMELEN];
	struct stat st;

	get_qfname(mnt, type, filename);
	debug(FL_DEBUG | FL_VERBOSE, "Data dumped.\nRenaming old quotafile to %s~\n", filename);
	if (stat(filename, &st) < 0) {	/* File doesn't exist? */
		if (errno == ENOENT) {
			debug(FL_DEBUG | FL_VERBOSE, "Old file not found.\n");
			goto rename_new;
		}
		die(2, "Error while searching for old quota file %s: %s\n", filename, strerror(errno));
	}
	/* Backup old file */
	strcpy(newfilename, filename);
	/* Make backingup safe */
	sstrncat(newfilename, "~", MAXFILENAMELEN);
	if (newfilename[strlen(newfilename) - 1] != '~')
		die(8, "Name of quota file too long. Contact %s.\n", MY_EMAIL);
	if (rename(filename, newfilename) < 0)
		die(2, "Can't rename old quotafile %s to %s: %s\n", filename, newfilename, strerror(errno));
	debug(FL_DEBUG | FL_VERBOSE, "Renaming new quotafile\n");
rename_new:
	/* Rename new file to right name */
	strcpy(newfilename, filename);
	sstrncat(newfilename, ".new", MAXFILENAMELEN);
	if (rename(newfilename, filename) < 0)
		die(2, "Can't rename new quotafile %s to name %s: %s\n", newfilename, filename, strerror(errno));
}

/*
 * Dump the quota info that we have in memory now to the appropriate
 * quota file. As quotafiles doesn't account to quotas we don't have to
 * bother about accounting new blocks for quota file
 */
void dump_to_file(char *dev, char *mnt, int type)
{
	struct check_dquot *dquot;
	struct dquot dumped;
	uint i;
	int kt = user2kernquota(type);

	debug(FL_DEBUG | FL_VERBOSE, "Dumping gathered data for %ss.\n", type2name(type));
	new_io(mnt, dev, type);
	info.dqi_bgrace = old_info[kt].dqi_bgrace;
	info.dqi_igrace = old_info[kt].dqi_igrace;
	info.dqi_flags = old_info[kt].dqi_flags;
	mark_quotafile_info_dirty();
	for (i = 0; i < DQUOTHASHSIZE; i++)
		for (dquot = dquot_hash[kt][i]; dquot; dquot = dquot->next) {
			dumped.dq_id = dquot->dq_id;
			dumped.dq_off = 0;
			memcpy(&dumped.dq_dqb, &dquot->dq_dqb, sizeof(struct mem_dqblk));
			/* Unset grace times if limit is not exceeded; if limit is not set, clear times too... */
			if (dumped.dq_dqb.dqb_bsoftlimit > toqb(dumped.dq_dqb.dqb_curspace))
				dumped.dq_dqb.dqb_btime = 0;
			if (dumped.dq_dqb.dqb_isoftlimit > dumped.dq_dqb.dqb_curinodes)
				dumped.dq_dqb.dqb_itime = 0;
			commit_dquot(&dumped, 1);
		}
	free_io();
	rename_files(mnt, type);
	if (!quotactl(QCMD(Q_GETINFO, kt), dev, 0, (void *)old_info + kt)) {	/* Is kernel quota turned on? */
		char filename[MAXFILENAMELEN];
		get_qfname(mnt, type, filename);
		if (quotactl(QCMD(Q_QUOTAOFF, kt), dev, 0, NULL) || quotactl(QCMD(Q_QUOTAON, kt), dev, 0, filename))
			fprintf(stderr, "Can't turn %s quotas on %s off and on: %s\nKernel won't know about changes quotacheck did.\n", type2name(type), mnt, strerror(errno)); 
	}
	else if (errno != ESRCH && errno != ENOSYS && errno != EINVAL)
		die(4, "Huh... Error while getting info for %s quota: %s\n", type2name(type), strerror(errno));
}

void check_dir(char *dev, char *mnt, char *fstype)
{
	struct stat st;
	struct statfs fsst;
	int remounted = 0;

	if (lstat(mnt, &st) < 0)
		die(2, "Can't stat mountpoint %s: %s\n", mnt, strerror(errno));
	if (statfs(mnt, &fsst) < 0)
		die(2, "Can't fsstat mountpoint %s: %s\n", mnt, strerror(errno));
	if (!S_ISDIR(st.st_mode))
		die(2, "Mountpoint %s isn't directory?!\n", mnt);
	cur_dev = st.st_dev;
	files_done = dirs_done = 0;
	if (check_type & QS_USRQUOTA) {
		process_file(dev, mnt, QS_USRQUOTA);
		add_to_quota(QS_USRQUOTA, st.st_ino, st.st_uid, st.st_gid, st.st_mode, st.st_nlink, st.st_blocks);
	}
	if (check_type & QS_GRPQUOTA) {
		process_file(dev, mnt, QS_GRPQUOTA);
		add_to_quota(QS_GRPQUOTA, st.st_ino, st.st_uid, st.st_gid, st.st_mode, st.st_nlink, st.st_blocks);
	}
	if (!(flags & FL_NOREMOUNT)) {
		/* Now we try to remount fs read-only to prevent races when scanning filesystem */
		if (mount(NULL, mnt, fstype, MS_MGC_VAL | MS_REMOUNT | MS_RDONLY, NULL) < 0 && !(flags & FL_FORCEREMOUNT)) {
			if (flags & FL_INTERACTIVE) {
				printf("Can't remount filesystem mounted on %s read-only. Counted values might not be right.\n", mnt);
				if (!ask_yn("Should I continue", 0))
					die(6, "As you wish... Finishing.\n");
			}
			else
				die(6, "Can't remount filesystem mounted on %s read-only so counted values might not be right.\n\
Please stop all programs writing to filesystem or use -F flag to force checking.\n", mnt);
		}
		else
			remounted = 1;
		debug(FL_DEBUG | FL_VERBOSE, "Filesystem remounted RO\n");
	}
	debug(FL_VERBOSE, "Scanning %s [%s] ", dev, mnt);
#if defined(EXT2_DIRECT)
	if (fsst.f_type == EXT2_SUPER_MAGIC) {
		ext2_direct_scan(dev);
	} else if (mnt) {
#else
	if (mnt) {
#endif
		scan_dir(mnt);
	}
	dirs_done++;
	if (flags & FL_VERBOSE)
		fputs("done\n", stderr);
	debug(FL_DEBUG | FL_VERBOSE, "Checked %d directories and %d files\n", dirs_done, files_done);
	if (remounted) {
		if (mount(NULL, mnt, fstype, MS_MGC_VAL | MS_REMOUNT, NULL) < 0)
			die(4, "Can't remount filesystem %s read-write. Can't write new quota files.\n", mnt);
		debug(FL_DEBUG | FL_VERBOSE, "Filesystem remounted RW.\n");
	}
	if (check_type & QS_USRQUOTA)
		dump_to_file(dev, mnt, QS_USRQUOTA);
	if (check_type & QS_GRPQUOTA)
		dump_to_file(dev, mnt, QS_GRPQUOTA);
	remove_list();
}

void check_all(void)
{
	char dev_name[MAXDEVLEN];
	char mnt_name[MAXFILENAMELEN];
	char fs_type[MAXFSTYPELEN];

	init_mount(&mntpoint, 1);
	if ((check_type = next_mount(type, dev_name, mnt_name, fs_type)))
		check_dir(dev_name, mnt_name, fs_type);
	else
		die(1, "Can't find mountpoint %s.\n", mntpoint);
	free_mount();
}

int main(int argcnt, char **argstr)
{
	parse_options(argcnt, argstr);

	check_all();
#ifdef DEBUG_MALLOC
	fprintf(stderr, "Allocated %d bytes memory\nFree'd %d bytes\nLost %d bytes\n", malloc_mem,
	free_mem, malloc_mem - free_mem);
#endif
	return 0;
}

